#! /usr/bin/env python3

import torch

from .generic_pair_loss import GenericPairLoss
from ..utils import loss_and_miner_utils as lmu


class GeneralizedLiftedStructureLoss(GenericPairLoss):
    # The 'generalized' lifted structure loss shown on page 4
    # of the "in defense of triplet loss" paper
    # https://arxiv.org/pdf/1703.07737.pdf
    def __init__(self, neg_margin, **kwargs):
        super().__init__(use_similarity=False, mat_based_loss=True, **kwargs)
        self.neg_margin = neg_margin

    def _compute_loss(self, mat, pos_mask, neg_mask):
        pos_loss = lmu.logsumexp(mat, keep_mask=pos_mask, add_one=False)
        neg_loss = lmu.logsumexp(self.neg_margin - mat, keep_mask=neg_mask, add_one=False)
        return torch.mean(torch.relu(pos_loss+neg_loss))



import torch
from .generic_pair_loss import GenericPairLoss
import numpy as np
from ..utils import loss_and_miner_utils as lmu

class TupletMarginLoss(GenericPairLoss):

    def __init__(self, margin, scale=64, **kwargs):
        super().__init__(**kwargs, use_similarity=True, mat_based_loss=False)
        self.margin = np.radians(margin)
        self.scale = scale
        self.add_to_recordable_attributes(list_of_names=["avg_pos_angle", "avg_neg_angle"])

    # pos_pairs and neg_pairs already represent cos(theta)
    def _compute_loss(self, pos_pairs, neg_pairs, indices_tuple):
        a1, _, a2, _ = indices_tuple

        if len(a1) > 0 and len(a2) > 0:
            pos_angles = torch.acos(pos_pairs)
            neg_angles = torch.acos(neg_pairs)
            self.avg_pos_angle = np.degrees(torch.mean(pos_angles).item())
            self.avg_neg_angle = np.degrees(torch.mean(neg_angles).item())
            pos_pairs = torch.cos(pos_angles - self.margin)
            pos_pairs = pos_pairs.unsqueeze(1)
            neg_pairs = neg_pairs.repeat(pos_pairs.size(0), 1)
            inside_exp = self.scale*(neg_pairs - pos_pairs)
            keep_mask = (a2.unsqueeze(0) == a1.unsqueeze(1)).float()
            return torch.mean(lmu.logsumexp(inside_exp, keep_mask=keep_mask, add_one=True, dim=1))
        return 0






from .base_weight_regularizer import BaseWeightRegularizer
import torch

class CenterInvariantRegularizer(BaseWeightRegularizer):
    def __init__(self, normalize_weights=False):
        super().__init__(normalize_weights)
        assert not self.normalize_weights, "normalize_weights must be False for CenterInvariantRegularizer"
  
    def compute_loss(self, weights):
        squared_weight_norms = self.weight_norms**2
        deviations_from_mean = squared_weight_norms - torch.mean(squared_weight_norms)
        return torch.mean((deviations_from_mean**2) / 4)


from .weight_regularizer_mixin import WeightRegularizerMixin
from .base_metric_loss_function import BaseMetricLossFunction
import torch
from ..utils import loss_and_miner_utils as lmu

# adapted from 
# https://github.com/tjddus9597/Proxy-Anchor-CVPR2020/blob/master/code/losses.py
# https://github.com/geonm/proxy-anchor-loss/blob/master/pytorch-proxy-anchor.py
# suggested in this issue: https://github.com/KevinMusgrave/pytorch-metric-learning/issues/32
class ProxyAnchorLoss(WeightRegularizerMixin, BaseMetricLossFunction):
    def __init__(self, num_classes, embedding_size, margin = 0.1, alpha = 32, **kwargs):
        super().__init__(**kwargs)
        self.proxies = torch.nn.Parameter(torch.randn(num_classes, embedding_size))
        torch.nn.init.kaiming_normal_(self.proxies, mode='fan_out')
        self.num_classes = num_classes
        self.margin = margin
        self.alpha = alpha
        
    def compute_loss(self, embeddings, labels, indices_tuple):
        miner_weights = lmu.convert_to_weights(indices_tuple, labels).unsqueeze(1)
        prox = torch.nn.functional.normalize(self.proxies, p=2, dim=1) if self.normalize_embeddings else self.proxies
        cos = lmu.sim_mat(embeddings, prox)

        if not self.normalize_embeddings:
            embedding_norms_mat = self.embedding_norms.unsqueeze(0)*torch.norm(prox, p=2, dim=1, keepdim=True)
            cos = cos / (embedding_norms_mat.t())

        pos_mask = torch.nn.functional.one_hot(labels, self.num_classes).float()
        neg_mask = 1 - pos_mask

        with_pos_proxies = torch.nonzero(torch.sum(pos_mask, dim=0) != 0).squeeze(1)

        pos_term = lmu.logsumexp(-self.alpha * (cos - self.margin), keep_mask=pos_mask*miner_weights, add_one=True, dim=0)
        neg_term = lmu.logsumexp(self.alpha * (cos + self.margin), keep_mask=neg_mask*miner_weights, add_one=True, dim=0)
        
        pos_term = torch.sum(pos_term) / len(with_pos_proxies)
        neg_term = torch.sum(neg_term) / self.num_classes
           
        return pos_term + neg_term + self.regularization_loss(self.proxies)



#! /usr/bin/env python3

from .large_margin_softmax_loss import LargeMarginSoftmaxLoss
import numpy as np
import torch

class ArcFaceLoss(LargeMarginSoftmaxLoss):
    """
    Implementation of https://arxiv.org/pdf/1801.07698.pdf
    """
    def __init__(self, scale=64, **kwargs):
        kwargs.pop("normalize_weights", None)
        super().__init__(scale=scale, normalize_weights=True, **kwargs)

    def init_margin(self):
        self.margin = np.radians(self.margin)

    def modify_cosine_of_target_classes(self, cosine_of_target_classes, *args):
        angles = self.get_angles(cosine_of_target_classes)
        return torch.cos(angles + self.margin)




#! /usr/bin/env python3

import numpy as np
from sklearn.metrics import normalized_mutual_info_score, adjusted_mutual_info_score
from . import stat_utils

def maybe_get_average_per_class(accuracy_per_sample, sample_labels, average_per_class):
    if average_per_class:
        unique_labels = np.unique(sample_labels)
        mask = sample_labels == unique_labels[None, :]
        acc_sum_per_class = np.sum(accuracy_per_sample[:, None]*mask, axis=0)
        mask_sum_per_class = np.sum(mask, axis=0)
        average_per_class = acc_sum_per_class / mask_sum_per_class
        return np.mean(average_per_class)
    return np.mean(accuracy_per_sample)

def get_relevance_mask(shape, gt_labels, embeddings_come_from_same_source, label_counts):
    relevance_mask = np.zeros(shape=shape, dtype=np.int)
    for k,v in label_counts.items():
        matching_rows = np.where(gt_labels==k)[0]
        max_column = v-1 if embeddings_come_from_same_source else v
        relevance_mask[matching_rows, :max_column] = 1
    return relevance_mask

def r_precision(knn_labels, gt_labels, embeddings_come_from_same_source, label_counts, average_per_class):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeddings_come_from_same_source, label_counts)
    matches_per_row = np.sum((knn_labels == gt_labels) * relevance_mask.astype(bool), axis=1) 
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    accuracy_per_sample = matches_per_row / max_possible_matches_per_row
    return maybe_get_average_per_class(accuracy_per_sample, gt_labels, average_per_class)

def mean_average_precision_at_r(knn_labels, gt_labels, embeddings_come_from_same_source, label_counts, average_per_class):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeddings_come_from_same_source, label_counts)
    num_samples, num_k = knn_labels.shape
    equality = (knn_labels == gt_labels) * relevance_mask.astype(bool)
    cumulative_correct = np.cumsum(equality, axis=1)
    k_idx = np.tile(np.arange(1, num_k + 1), (num_samples, 1))
    precision_at_ks = (cumulative_correct * equality) / k_idx
    summed_precision_per_row = np.sum(precision_at_ks * relevance_mask, axis=1)
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    accuracy_per_sample = summed_precision_per_row / max_possible_matches_per_row
    return maybe_get_average_per_class(accuracy_per_sample, gt_labels, average_per_class)

def precision_at_k(knn_labels, gt_labels, k, average_per_class):
    curr_knn_labels = knn_labels[:, :k]
    accuracy_per_sample = np.sum(curr_knn_labels == gt_labels, axis=1) / k
    return maybe_get_average_per_class(accuracy_per_sample, gt_labels, average_per_class)

def get_label_counts(reference_labels):
    unique_labels, label_counts = np.unique(reference_labels, return_counts=True)
    num_k = min(1023, int(np.max(label_counts))) # faiss can only do a max of k=1024, and we have to do k+1
    return {k:v for k,v in zip(unique_labels, label_counts)}, num_k

def get_lone_query_labels(query_labels, reference_labels, reference_label_counts, embeddings_come_from_same_source):
    if embeddings_come_from_same_source:
        return np.array([k for k,v in reference_label_counts.items() if v <= 1])
    else:
        return np.setdiff1d(query_labels, reference_labels)

class AccuracyCalculator:
    def __init__(self, include=(), exclude=(), average_per_class=False):
        self.function_keyword = "calculate_"
        function_names = [x for x in dir(self) if x.startswith(self.function_keyword)]
        metrics = [x.replace(self.function_keyword, "", 1) for x in function_names]
        self.original_function_dict = {x:getattr(self, y) for x,y in zip(metrics, function_names)}
        self.check_primary_metrics(include, exclude)
        self.original_function_dict = self.get_function_dict(include, exclude)
        self.curr_function_dict = self.get_function_dict()
        self.average_per_class = average_per_class

    def get_function_dict(self, include=(), exclude=()):
        if len(include) == 0:
            include = list(self.original_function_dict.keys())
        included_metrics = [k for k in include if k not in exclude] 
        return {k:v for k,v in self.original_function_dict.items() if k in included_metrics}

    def get_curr_metrics(self):
        return [k for k in self.curr_function_dict.keys()]

    def requires_clustering(self):
        return ["NMI", "AMI"]

    def requires_knn(self):
        return ["precision_at_1", "mean_average_precision_at_r", "r_precision"]

    def get_cluster_labels(self, query, query_labels, **kwargs):
        num_clusters = len(np.unique(query_labels.flatten()))
        return stat_utils.run_kmeans(query, num_clusters)

    def calculate_NMI(self, query_labels, cluster_labels, **kwargs):
        return normalized_mutual_info_score(query_labels, cluster_labels)

    def calculate_AMI(self, query_labels, cluster_labels, **kwargs):
        return adjusted_mutual_info_score(query_labels, cluster_labels)

    def calculate_precision_at_1(self, knn_labels, query_labels, not_lone_query_idx, **kwargs):
        knn_labels, query_labels = knn_labels[not_lone_query_idx], query_labels[not_lone_query_idx]
        return precision_at_k(knn_labels, query_labels[:, None], 1, self.average_per_class)
        
    def calculate_mean_average_precision_at_r(self, knn_labels, query_labels, not_lone_query_idx, embeddings_come_from_same_source, label_counts, **kwargs):
        knn_labels, query_labels = knn_labels[not_lone_query_idx], query_labels[not_lone_query_idx]
        return mean_average_precision_at_r(knn_labels, query_labels[:, None], embeddings_come_from_same_source, label_counts, self.average_per_class)

    def calculate_r_precision(self, knn_labels, query_labels, not_lone_query_idx, embeddings_come_from_same_source, label_counts, **kwargs):
        knn_labels, query_labels = knn_labels[not_lone_query_idx], query_labels[not_lone_query_idx]
        return r_precision(knn_labels, query_labels[:, None], embeddings_come_from_same_source, label_counts, self.average_per_class)

    def get_accuracy(self, query, reference, query_labels, reference_labels, embeddings_come_from_same_source, include=(), exclude=()):
        embeddings_come_from_same_source = embeddings_come_from_same_source or (query is reference)

        self.curr_function_dict = self.get_function_dict(include, exclude)

        kwargs = {"query": query, 
                "reference": reference,
                "query_labels": query_labels,
                "reference_labels": reference_labels,
                "embeddings_come_from_same_source": embeddings_come_from_same_source}

        if any(x in self.requires_knn() for x in self.get_curr_metrics()):
            label_counts, num_k = get_label_counts(reference_labels)
            knn_indices = stat_utils.get_knn(reference, query, num_k, embeddings_come_from_same_source)
            knn_labels = reference_labels[knn_indices]
            lone_query_labels = get_lone_query_labels(query_labels, reference_labels, label_counts, embeddings_come_from_same_source)
            not_lone_query_idx = ~np.isin(query_labels, lone_query_labels)
            kwargs["label_counts"] = label_counts
            kwargs["knn_labels"] = knn_labels
            kwargs["lone_query_labels"] = lone_query_labels
            kwargs["not_lone_query_idx"] = not_lone_query_idx

        if any(x in self.requires_clustering() for x in self.get_curr_metrics()):
            kwargs["cluster_labels"] = self.get_cluster_labels(**kwargs)                

        return self._get_accuracy(self.curr_function_dict, **kwargs)

    def _get_accuracy(self, function_dict, **kwargs):
        return {k:v(**kwargs) for k,v in function_dict.items()}

    def check_primary_metrics(calc, include=(), exclude=()):
        primary_metrics = list(calc.original_function_dict.keys())
        for met in [include, exclude]:
            if not isinstance(met, (tuple, list)):
                raise TypeError("Arguments must be of type tuple, not {}.".format(type(met)))
            if not set(met).issubset(set(primary_metrics)):
                raise ValueError("Primary metrics must be one or more of: {}.".format(primary_metrics))

    def description(self):
        return "avg_per_class" if self.average_per_class else ""



from .large_margin_softmax_loss import LargeMarginSoftmaxLoss

class SphereFaceLoss(LargeMarginSoftmaxLoss):
	# implementation of https://arxiv.org/pdf/1704.08063.pdf
    def __init__(self, **kwargs):
        kwargs.pop("normalize_weights", None)
        super().__init__(normalize_weights=True, **kwargs)


#! /usr/bin/env python3

import torch

from .generic_pair_loss import GenericPairLoss
from ..utils import common_functions as c_f, loss_and_miner_utils as lmu

class MultiSimilarityLoss(GenericPairLoss):
    """
    modified from https://github.com/MalongTech/research-ms-loss/
    Args:
        alpha: The exponential weight for positive pairs
        beta: The exponential weight for negative pairs
        base: The shift in the exponent applied to both positive and negative pairs
    """
    def __init__(self, alpha, beta, base=0.5, **kwargs):
        super().__init__(use_similarity=True, mat_based_loss=True, **kwargs)
        self.alpha = alpha
        self.beta = beta
        self.base = base

    def _compute_loss(self, mat, pos_mask, neg_mask):
        pos_loss = (1.0/self.alpha) * lmu.logsumexp(-self.alpha * (mat - self.base), keep_mask=pos_mask, add_one=True)
        neg_loss = (1.0/self.beta) * lmu.logsumexp(self.beta * (mat - self.base), keep_mask=neg_mask, add_one=True)
        return torch.mean(pos_loss + neg_loss)


#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import torch 
import torch.nn.functional as F 
from ..utils import loss_and_miner_utils as lmu 


class CircleLoss(BaseMetricLossFunction):
    """
    Circle loss for pairwise labels only. Support for class-level labels will be added 
    in the future.
    
    Args:
    m:  The relaxation factor that controls the radious of the decision boundary.
    gamma: The scale factor that determines the largest scale of each similarity score.

    According to the paper, the suggested default values of m and gamma are:

    Face Recognition: m = 0.25, gamma = 256
    Person Reidentification: m = 0.25, gamma = 256
    Fine-grained Image Retrieval: m = 0.4, gamma = 80

    By default, we set m = 0.4 and gamma = 80
    """
    def __init__(
        self, 
        m=0.4,
        gamma=80,
        triplets_per_anchor='all',
        **kwargs
    ):
        super(CircleLoss, self).__init__(**kwargs)
        self.m = m 
        self.gamma = gamma 
        self.triplets_per_anchor = triplets_per_anchor
        self.add_to_recordable_attributes(list_of_names=["num_unique_anchors", "num_triplets"])
        self.soft_plus = torch.nn.Softplus(beta=1)

        assert self.normalize_embeddings, "Embeddings must be normalized in circle loss!"

    def compute_loss(self, embeddings, labels, indices_tuple):
        indices_tuple = lmu.convert_to_triplets(indices_tuple, labels, t_per_anchor=self.triplets_per_anchor)
        anchor_idx, positive_idx, negative_idx = indices_tuple 
        self.num_triplets = len(anchor_idx)
        if self.num_triplets == 0:
            self.num_unique_anchors = 0
            return 0
        anchors, positives, negatives = embeddings[anchor_idx], embeddings[positive_idx], embeddings[negative_idx]
        
        # compute cosine similarities
        # since embeddings are normalized, we only need to compute dot product 
        sp = torch.sum(anchors * positives, dim=1)
        sn = torch.sum(anchors * negatives, dim=1)
        
        # compute some constants
        loss = 0.
        op = 1 + self.m 
        on = -self.m
        delta_p = 1-self.m 
        delta_n = self.m 

        # find unique anchor index 
        # for each unique anchor index, we have (sp1, sp2, ..., spK) (sn1, sn2, ..., snL)
        unique_anchor_idx = torch.unique(anchor_idx)
        self.num_unique_anchors = len(unique_anchor_idx)

        for anchor in unique_anchor_idx:
            mask = anchor_idx == anchor 
            sp_for_this_anchor = sp[mask]
            sn_for_this_anchor = sn[mask]
            alpha_p = torch.clamp(op - sp_for_this_anchor.detach(), min=0.)
            alpha_n = torch.clamp(sn_for_this_anchor.detach() - on, min=0.)

            logit_p = -self.gamma * alpha_p * (sp_for_this_anchor - delta_p)
            logit_n = self.gamma * alpha_n * (sn_for_this_anchor - delta_n)

            loss_for_this_anchor = self.soft_plus(
                torch.logsumexp(logit_n, dim=0) + torch.logsumexp(logit_p, dim=0))
            loss += loss_for_this_anchor
        
        loss /= len(unique_anchor_idx)
        return loss









import torch
from ..utils import common_functions as c_f, loss_and_miner_utils as lmu

class CrossBatchMemory(torch.nn.Module):
    def __init__(self, loss, embedding_size, memory_size=1024, miner=None):
        super().__init__()
        self.loss = loss
        self.miner = miner
        self.memory_size = memory_size
        self.embedding_memory = torch.zeros(self.memory_size, embedding_size)
        self.label_memory = torch.zeros(self.memory_size).long()
        self.has_been_filled = False
        self.queue_idx = 0

    def forward(self, embeddings, labels, input_indices_tuple=None):
        assert embeddings.size(0) <= self.embedding_memory.size(0)
        batch_size = embeddings.size(0)
        labels = labels.to(embeddings.device)
        self.embedding_memory = self.embedding_memory.to(embeddings.device)
        self.label_memory = self.label_memory.to(labels.device)
        self.add_to_memory(embeddings, labels, batch_size)
        
        if not self.has_been_filled:
            E_mem = self.embedding_memory[:self.queue_idx]
            L_mem = self.label_memory[:self.queue_idx] 
        else:
            E_mem = self.embedding_memory
            L_mem = self.label_memory

        indices_tuple = self.create_indices_tuple(batch_size, embeddings, labels, E_mem, L_mem, input_indices_tuple)
        combined_embeddings = torch.cat([embeddings, E_mem], dim=0)
        combined_labels = torch.cat([labels, L_mem], dim=0)
        loss = self.loss(combined_embeddings, combined_labels, indices_tuple)
        return loss

    def add_to_memory(self, embeddings, labels, batch_size):
        end_idx = ((self.queue_idx + batch_size - 1) % (self.memory_size)) + 1

        if end_idx > self.queue_idx:
            self.embedding_memory[self.queue_idx:end_idx] = embeddings.detach()
            self.label_memory[self.queue_idx:end_idx] = labels.detach()            
        else:
            se = self.memory_size-self.queue_idx
            self.embedding_memory[self.queue_idx:] = embeddings[:se].detach()
            self.embedding_memory[:end_idx] = embeddings[se:].detach()
            self.label_memory[self.queue_idx:] = labels[:se].detach()
            self.label_memory[:end_idx] = labels[se:].detach()
            

        prev_queue_idx = self.queue_idx
        self.queue_idx = (self.queue_idx + batch_size) % self.memory_size

        if (not self.has_been_filled) and (self.queue_idx <= prev_queue_idx):
            self.has_been_filled = True


    def create_indices_tuple(self, batch_size, embeddings, labels, E_mem, L_mem, input_indices_tuple):
        if self.miner:
        	indices_tuple = self.miner(embeddings, labels, E_mem, L_mem)
        else:
        	indices_tuple = lmu.get_all_pairs_indices(labels, L_mem)
        
        indices_tuple = c_f.shift_indices_tuple(indices_tuple, batch_size)

        if input_indices_tuple is not None:
            if len(input_indices_tuple) == 3 and len(indices_tuple) == 4:
                input_indices_tuple = lmu.convert_to_pairs(input_indices_tuple, labels)
            elif len(input_indices_tuple) == 4 and len(indices_tuple) == 3:
                input_indices_tuple = lmu.convert_to_triplets(input_indices_tuple, labels)
            indices_tuple = tuple([torch.cat([x,y.to(x.device)], dim=0) for x,y in zip(indices_tuple, input_indices_tuple)])

        return indices_tuple